from collections import defaultdict
from typing import Optional

from bec_lib.endpoints import MessageEndpoints
from pydantic import BaseModel, Field
from qtpy.QtCore import Property, Signal, Slot
from qtpy.QtWidgets import (
    QApplication,
    QComboBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.qt_utils.error_popups import SafeSlot
from bec_widgets.utils import ConnectionConfig
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.widgets.scan_control.scan_group_box import ScanGroupBox
from bec_widgets.widgets.stop_button.stop_button import StopButton
from bec_widgets.widgets.toggle.toggle import ToggleSwitch


class ScanParameterConfig(BaseModel):
    name: str
    args: Optional[list] = Field(None)
    kwargs: Optional[dict] = Field(None)


class ScanControlConfig(ConnectionConfig):
    default_scan: Optional[str] = Field(None)
    allowed_scans: Optional[list] = Field(None)
    scans: Optional[dict[str, ScanParameterConfig]] = defaultdict(dict)


class ScanControl(BECWidget, QWidget):

    ICON_NAME = "tune"

    scan_started = Signal()
    scan_selected = Signal(str)

    def __init__(
        self,
        parent=None,
        client=None,
        config: ScanControlConfig | dict | None = None,
        gui_id: str | None = None,
        allowed_scans: list | None = None,
        default_scan: str | None = None,
    ):

        if config is None:
            config = ScanControlConfig(
                widget_class=self.__class__.__name__, allowed_scans=allowed_scans
            )
        super().__init__(client=client, gui_id=gui_id, config=config)
        QWidget.__init__(self, parent=parent)

        # Client from BEC + shortcuts to device manager and scans
        self.get_bec_shortcuts()

        # Main layout
        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(5, 5, 5, 5)
        self.arg_box = None
        self.kwarg_boxes = []
        self.expert_mode = False  # TODO implement in the future versions
        self.previous_scan = None
        self.last_scan_found = None

        # Widget Default Parameters
        self.config.default_scan = default_scan
        self.config.allowed_scans = allowed_scans

        # Create and set main layout
        self._init_UI()

    def _init_UI(self):
        """
        Initializes the UI of the scan control widget. Create the top box for scan selection and populate scans to main combobox.
        """

        # Scan selection group box
        self.scan_selection_group = self.create_scan_selection_group()
        self.scan_selection_group.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        self.layout.addWidget(self.scan_selection_group)

        # Default scan from config
        if self.config.default_scan is not None:
            self.comboBox_scan_selection.setCurrentText(self.config.default_scan)

        # Connect signals
        self.comboBox_scan_selection.view().pressed.connect(self.save_current_scan_parameters)
        self.comboBox_scan_selection.currentIndexChanged.connect(self.on_scan_selection_changed)
        self.button_run_scan.clicked.connect(self.run_scan)

        # Add bundle button
        self.button_add_bundle = QPushButton("Add Bundle")
        self.button_add_bundle.setVisible(False)
        # Remove bundle button
        self.button_remove_bundle = QPushButton("Remove Bundle")
        self.button_remove_bundle.setVisible(False)

        bundle_layout = QHBoxLayout()
        bundle_layout.addWidget(self.button_add_bundle)
        bundle_layout.addWidget(self.button_remove_bundle)
        self.layout.addLayout(bundle_layout)

        self.button_add_bundle.clicked.connect(self.add_arg_bundle)
        self.button_remove_bundle.clicked.connect(self.remove_arg_bundle)

        self.scan_selected.connect(self.scan_select)

        # Initialize scan selection
        self.populate_scans()

    def create_scan_selection_group(self) -> QGroupBox:
        """
        Creates the scan selection group box with combobox to select the scan and start/stop button.

        Returns:
            QGroupBox: Group box containing the scan selection widgets.
        """

        scan_selection_group = QGroupBox("Scan Selection", self)
        self.scan_selection_layout = QVBoxLayout(scan_selection_group)
        self.comboBox_scan_selection = QComboBox(scan_selection_group)

        # Buttons
        self.button_layout = QHBoxLayout()
        ## Run button
        self.button_run_scan = QPushButton("Start", scan_selection_group)
        self.button_run_scan.setStyleSheet("background-color:  #559900; color: white")
        ## Stop button
        self.button_stop_scan = StopButton(parent=scan_selection_group)
        ## Add buttons to layout
        self.button_layout.addWidget(self.button_run_scan)
        self.button_layout.addWidget(self.button_stop_scan)

        # Label to reload the last scan parameters
        self.toggle_layout = QHBoxLayout()
        ## Label
        self.last_scan_label = QLabel("Restore last scan parameters", scan_selection_group)
        ## Switch toggle button
        self.toggle = ToggleSwitch(parent=scan_selection_group, checked=False)
        self.toggle.enabled.connect(self.request_last_executed_scan_parameters)
        ## Add label and switch to layout
        self.toggle_layout.addWidget(self.last_scan_label)
        self.toggle_layout.addWidget(self.toggle)

        # Add widgets to layout
        self.scan_selection_layout.addWidget(self.comboBox_scan_selection)
        self.scan_selection_layout.addLayout(self.button_layout)
        self.scan_selection_layout.addLayout(self.toggle_layout)

        return scan_selection_group

    def populate_scans(self):
        """Populates the scan selection combo box with available scans from BEC session."""
        self.available_scans = self.client.connector.get(
            MessageEndpoints.available_scans()
        ).resource
        if self.config.allowed_scans is None:
            supported_scans = ["ScanBase", "SyncFlyScanBase", "AsyncFlyScanBase"]
            allowed_scans = [
                scan_name
                for scan_name, scan_info in self.available_scans.items()
                if scan_info["base_class"] in supported_scans and len(scan_info["gui_config"]) > 0
            ]

        else:
            allowed_scans = self.config.allowed_scans
        self.comboBox_scan_selection.addItems(allowed_scans)

    def on_scan_selection_changed(self, index: int):
        """Callback for scan selection combo box"""
        selected_scan_name = self.comboBox_scan_selection.currentText()
        self.scan_selected.emit(selected_scan_name)
        self.request_last_executed_scan_parameters()
        self.restore_scan_parameters(selected_scan_name)

    @Slot()
    def request_last_executed_scan_parameters(self):
        """
        Requests the last executed scan parameters from BEC and restores them to the scan control widget.
        """
        enabled = self.toggle.checked
        current_scan = self.comboBox_scan_selection.currentText()
        if enabled:
            history = self.client.connector.lrange(MessageEndpoints.scan_queue_history(), 0, -1)

            for scan in history:
                scan_name = scan.content["info"]["request_blocks"][-1]["msg"].content["scan_type"]
                if scan_name == current_scan:
                    args_dict = scan.content["info"]["request_blocks"][-1]["msg"].content[
                        "parameter"
                    ]["args"]
                    args_list = []
                    for key, value in args_dict.items():
                        args_list.append(key)
                        args_list.extend(value)
                    if len(args_list) > 1 and self.arg_box is not None:
                        self.arg_box.set_parameters(args_list)
                    kwargs = scan.content["info"]["request_blocks"][-1]["msg"].content["parameter"][
                        "kwargs"
                    ]
                    if kwargs and self.kwarg_boxes:
                        for box in self.kwarg_boxes:
                            box.set_parameters(kwargs)
                    self.last_scan_found = True
                    break
                else:
                    self.last_scan_found = False
        else:
            self.last_scan_found = False

    @Property(str)
    def current_scan(self):
        """Returns the scan name for the currently selected scan."""
        return self.comboBox_scan_selection.currentText()

    @current_scan.setter
    def current_scan(self, scan_name: str):
        """Sets the current scan to the given scan name.

        Args:
            scan_name(str): Name of the scan to set as current.
        """
        if scan_name not in self.available_scans:
            return
        self.comboBox_scan_selection.setCurrentText(scan_name)

    @Slot(str)
    def set_current_scan(self, scan_name: str):
        """Slot for setting the current scan to the given scan name.

        Args:
            scan_name(str): Name of the scan to set as current.
        """
        self.current_scan = scan_name

    @Property(bool)
    def hide_scan_remember_toggle(self):
        """Property to hide the scan remember toggle."""
        return not self.toggle.isVisible()

    @hide_scan_remember_toggle.setter
    def hide_scan_remember_toggle(self, hide: bool):
        """Setter for the hide_scan_remember_toggle property.

        Args:
            hide(bool): Hide or show the scan remember toggle.
        """
        self.show_scan_remember_toggle(not hide)

    @Slot(bool)
    def show_scan_remember_toggle(self, show: bool):
        """Shows or hides the scan control buttons."""
        self.toggle.setVisible(show)
        self.last_scan_label.setVisible(show)

        show_group = show or self.button_run_scan.isVisible()
        self.scan_selection_group.setVisible(show_group)

    @Property(bool)
    def hide_scan_control_buttons(self):
        """Property to hide the scan control buttons."""
        return not self.button_run_scan.isVisible()

    @hide_scan_control_buttons.setter
    def hide_scan_control_buttons(self, hide: bool):
        """Setter for the hide_scan_control_buttons property.

        Args:
            hide(bool): Hide or show the scan control buttons.
        """
        self.show_scan_control_buttons(not hide)

    @Slot(bool)
    def show_scan_control_buttons(self, show: bool):
        """Shows or hides the scan control buttons."""
        self.button_run_scan.setVisible(show)
        self.button_stop_scan.setVisible(show)

        show_group = show or self.button_run_scan.isVisible()
        self.scan_selection_group.setVisible(show_group)

    @Property(bool)
    def hide_scan_selection_combobox(self):
        """Property to hide the scan selection combobox."""
        return not self.comboBox_scan_selection.isVisible()

    @hide_scan_selection_combobox.setter
    def hide_scan_selection_combobox(self, hide: bool):
        """Setter for the hide_scan_selection_combobox property.

        Args:
            hide(bool): Hide or show the scan selection combobox.
        """
        self.show_scan_selection_combobox(not hide)

    @Slot(bool)
    def show_scan_selection_combobox(self, show: bool):
        """Shows or hides the scan selection combobox."""
        self.comboBox_scan_selection.setVisible(show)

        show_group = show or self.button_run_scan.isVisible()
        self.scan_selection_group.setVisible(show_group)

    @Slot(str)
    def scan_select(self, scan_name: str):
        """
        Slot for scan selection. Updates the scan control layout based on the selected scan.

        Args:
            scan_name(str): Name of the selected scan.
        """
        self.reset_layout()
        selected_scan_info = self.available_scans.get(scan_name, {})

        gui_config = selected_scan_info.get("gui_config", {})
        self.arg_group = gui_config.get("arg_group", None)
        self.kwarg_groups = gui_config.get("kwarg_groups", None)

        show_bundle_buttons = bool(self.arg_group["arg_inputs"])

        self._show_bundle_buttons(show_bundle_buttons)

        if show_bundle_buttons:
            self.add_arg_group(self.arg_group)
        if len(self.kwarg_groups) > 0:
            self.add_kwargs_boxes(self.kwarg_groups)

        self.update()
        self.adjustSize()

    def _show_bundle_buttons(self, show: bool):
        """Shows or hides the bundle buttons based on the show argument.

        Args:
            show(bool): Show or hide the bundle buttons.
        """
        self.button_add_bundle.setVisible(show)
        self.button_remove_bundle.setVisible(show)

    @Property(bool)
    def hide_bundle_buttons(self):
        """Property to hide the bundle buttons."""
        return not self.button_add_bundle.isVisible()

    @hide_bundle_buttons.setter
    def hide_bundle_buttons(self, hide: bool):
        """Setter for the hide_bundle_buttons property.

        Args:
            hide(bool): Hide or show the bundle buttons.
        """
        self._show_bundle_buttons(not hide)

    def add_kwargs_boxes(self, groups: list):
        """
        Adds the given gui_groups to the scan control layout.

        Args:
            groups(list): List of dictionaries containing the gui_group information.
        """
        for group in groups:
            box = ScanGroupBox(box_type="kwargs", config=group)
            box.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
            self.layout.addWidget(box)
            self.kwarg_boxes.append(box)

    def add_arg_group(self, group: dict):
        """
        Adds the given gui_groups to the scan control layout.

        Args:
        """
        self.arg_box = ScanGroupBox(box_type="args", config=group)
        self.arg_box.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        self.layout.addWidget(self.arg_box)

    @Slot()
    def add_arg_bundle(self):
        """Adds a new argument bundle to the scan control layout."""
        self.arg_box.add_widget_bundle()

    @Slot()
    def remove_arg_bundle(self):
        """Removes the last argument bundle from the scan control layout."""
        self.arg_box.remove_widget_bundle()

    def reset_layout(self):
        """Clears the scan control layout from GuiGroups and ArgGroups boxes."""
        if self.arg_box is not None:
            self.layout.removeWidget(self.arg_box)
            self.arg_box.deleteLater()
            self.arg_box = None
        if self.kwarg_boxes != []:
            self.remove_kwarg_boxes()

    def remove_kwarg_boxes(self):
        for box in self.kwarg_boxes:
            self.layout.removeWidget(box)
            box.deleteLater()
        self.kwarg_boxes = []

    def get_scan_parameters(self, bec_object: bool = True):
        """
        Returns the scan parameters for the selected scan.

        Args:
            bec_object(bool): If True, returns the BEC object for the scan parameters such as device objects.
        """
        args = []
        kwargs = {}
        if self.arg_box is not None:
            args = self.arg_box.get_parameters(bec_object)
        for box in self.kwarg_boxes:
            box_kwargs = box.get_parameters(bec_object)
            kwargs.update(box_kwargs)
        return args, kwargs

    def restore_scan_parameters(self, scan_name: str):
        """
        Restores the scan parameters for the given scan name

        Args:
            scan_name(str): Name of the scan to restore the parameters for.
        """
        if self.last_scan_found is True:
            return
        scan_params = self.config.scans.get(scan_name, None)
        if scan_params is None and self.previous_scan is None:
            return

        if scan_params is None and self.previous_scan is not None:
            previous_scan_params = self.config.scans.get(self.previous_scan, None)
            self._restore_kwargs(previous_scan_params.kwargs)
            return

        if scan_params.args is not None and self.arg_box is not None:
            self.arg_box.set_parameters(scan_params.args)

        self._restore_kwargs(scan_params.kwargs)

    def _restore_kwargs(self, scan_kwargs: dict):
        """Restores the kwargs for the given scan parameters."""
        if scan_kwargs is not None and self.kwarg_boxes is not None:
            for box in self.kwarg_boxes:
                box.set_parameters(scan_kwargs)

    def save_current_scan_parameters(self):
        """Saves the current scan parameters to the scan control config for further use."""
        scan_name = self.comboBox_scan_selection.currentText()
        self.previous_scan = scan_name
        args, kwargs = self.get_scan_parameters(False)
        scan_params = ScanParameterConfig(name=scan_name, args=args, kwargs=kwargs)
        self.config.scans[scan_name] = scan_params

    @SafeSlot(popup_error=True)
    def run_scan(self):
        """Starts the selected scan with the given parameters."""
        self.scan_started.emit()
        args, kwargs = self.get_scan_parameters()
        scan_function = getattr(self.scans, self.comboBox_scan_selection.currentText())
        if callable(scan_function):
            scan_function(*args, **kwargs)

    def cleanup(self):
        """Cleanup the scan control widget."""
        self.button_stop_scan.cleanup()
        if self.arg_box:
            for widget in self.arg_box.widgets:
                if hasattr(widget, "cleanup"):
                    widget.cleanup()
        for kwarg_box in self.kwarg_boxes:
            for widget in kwarg_box.widgets:
                if hasattr(widget, "cleanup"):
                    widget.cleanup()
        super().cleanup()


# Application example
if __name__ == "__main__":  # pragma: no cover
    from bec_widgets.utils.colors import set_theme

    app = QApplication([])
    scan_control = ScanControl()

    set_theme("auto")
    window = scan_control
    window.show()
    app.exec()
